# Generated by Django 4.2 on 2023-04-16 21:49

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('auth', '0012_alter_user_first_name_max_length'),
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('safety', '0003_alter_objectpermission_options'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='objectpermission',
            unique_together=set(),
        ),
        migrations.AlterField(
            model_name='objectpermission',
            name='object_ct',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='object_of',
                                    to='contenttypes.contenttype', verbose_name='Target Content Type'),
        ),
        migrations.AlterField(
            model_name='objectpermission',
            name='object_id',
            field=models.IntegerField(verbose_name='Object ID'),
        ),
        migrations.AlterField(
            model_name='objectpermission',
            name='to_ct',
            field=models.ForeignKey(limit_choices_to={'model__in': ('user', 'group')},
                                    on_delete=django.db.models.deletion.CASCADE, related_name='entity_of',
                                    to='contenttypes.contenttype', verbose_name='Content Type'),
        ),
        migrations.AlterUniqueTogether(
            name='objectpermission',
            unique_together={('to_ct', 'permission', 'object_ct', 'object_id')},
        ),
        migrations.CreateModel(
            name='PermissionGroup',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, verbose_name='Name')),
                ('target_id', models.IntegerField(verbose_name='Target ID')),
                ('permissions', models.ManyToManyField(to='auth.permission', verbose_name='Permissions')),
                ('target_ct',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype',
                                   verbose_name='Target Content Type')),
                ('users', models.ManyToManyField(to=settings.AUTH_USER_MODEL, verbose_name='Users')),
            ],
            options={
                'verbose_name': 'Permission Group',
                'verbose_name_plural': 'Permission Groups',
            },
        ),
    ]



# Generated by Django 4.2 on 2023-04-22 15:36

from django.conf import settings
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('auth', '0012_alter_user_first_name_max_length'),
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('safety', '0007_alter_objectpermission_options'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='PermissionGroup',
            new_name='ObjectGroup',
        ),
    ]



# Generated by Django 4.2 on 2023-04-18 23:07

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('auth', '0012_alter_user_first_name_max_length'),
        ('safety', '0004_alter_objectpermission_unique_together_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='permissiongroup',
            name='target_id',
            field=models.IntegerField(null=True, verbose_name='Target ID'),
        ),
        migrations.AlterUniqueTogether(
            name='objectpermission',
            unique_together={('to_ct', 'to_id', 'permission', 'object_ct', 'object_id')},
        ),
    ]



# Generated by Django 4.2 on 2023-04-16 19:49

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('auth', '0012_alter_user_first_name_max_length'),
    ]

    operations = [
        migrations.CreateModel(
            name='ObjectPermission',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('to_id', models.CharField(max_length=255, verbose_name='Target ID')),
                ('object_id', models.CharField(max_length=255, verbose_name='Object ID')),
                ('object_ct',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype',
                                   verbose_name='Content Type')),
                ('permission', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='auth.permission',
                                                 verbose_name='Permission')),
                ('to_ct', models.ForeignKey(limit_choices_to={'model__in': ('user', 'group')},
                                            on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype',
                                            verbose_name='Content Type')),
            ],
            options={
                'verbose_name': 'User object permission',
                'verbose_name_plural': 'User object permissions',
                'unique_together': {('to_ct', 'permission', 'object_ct', 'object_id')},
            },
        ),
    ]



# Generated by Django 4.2 on 2023-04-22 15:06

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('safety', '0005_alter_permissiongroup_target_id_and_more'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='objectpermission',
            options={'verbose_name': 'User object permission', 'verbose_name_plural': 'User object permissions'},
        ),
    ]



# Generated by Django 4.2 on 2023-04-16 21:27

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('auth', '0012_alter_user_first_name_max_length'),
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('safety', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='objectpermission',
            name='object_ct',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype',
                                    verbose_name='Target Content Type'),
        ),
        migrations.AlterField(
            model_name='objectpermission',
            name='object_id',
            field=models.IntegerField(verbose_name='Object ID'),
        ),
    ]



from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.db import models
from django.utils.translation import gettext_lazy as _


class AbstractObjectPermission(models.Model):
    """
    Contains the fields that are common to both user and group object permissions.
    """

    permission = models.ForeignKey('auth.Permission', on_delete=models.CASCADE, verbose_name=_('Permission'))

    to = GenericForeignKey('to_ct', 'to_id')
    to_id = models.CharField(_('Target ID'), max_length=255)
    to_ct = models.ForeignKey('contenttypes.ContentType', on_delete=models.CASCADE, verbose_name=_('Content Type'),
                              limit_choices_to={'model__in': ('user', 'group')}, related_name='entity_of')

    object_id = models.IntegerField(_('Object ID'))
    object_ct = models.ForeignKey('contenttypes.ContentType', on_delete=models.CASCADE,
                                  verbose_name=_('Target Content Type'), related_name='object_of')
    object = GenericForeignKey('object_ct', 'object_id')

    class Meta:
        unique_together = (('to_ct', 'to_id', 'permission', 'object_ct', 'object_id'),)
        abstract = True

    def __str__(self):
        return f'{self.to} has {self.permission} on {self.object}'


class ObjectPermission(AbstractObjectPermission):
    class Meta(AbstractObjectPermission.Meta):
        verbose_name = _('User Object Permission')
        verbose_name_plural = _('User Object Permissions')

        swappable = 'SAFETY_OBJECT_PERMISSION_MODEL'


class AbstractObjectGroup(models.Model):
    """
    A group of permissions for an object.
    """

    name = models.CharField(_('Name'), max_length=255)
    users = models.ManyToManyField(settings.AUTH_USER_MODEL, verbose_name=_('Users'))
    permissions = models.ManyToManyField('auth.Permission', verbose_name=_('Permissions'))

    target = GenericForeignKey('target_ct', 'target_id')
    target_id = models.IntegerField(_('Target ID'), null=True)
    target_ct = models.ForeignKey('contenttypes.ContentType', on_delete=models.CASCADE,
                                  verbose_name=_('Target Content Type'))

    class Meta:
        abstract = True

    def __str__(self):
        return self.name


class ObjectGroup(AbstractObjectGroup):
    class Meta:
        verbose_name = _('Permission Group')
        verbose_name_plural = _('Permission Groups')

        swappable = 'SAFETY_OBJECT_GROUP_MODEL'



from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.test import TransactionTestCase
from django_fake_model import models as f

from safety.shortcuts import set_perm, has_perm, lift_perm, create_object_group, delete_object_group, \
    add_user_to_object_group, remove_user_from_object_group


class FakePost(f.FakeModel):
    title = models.CharField(max_length=100)
    content = models.TextField()

    class Meta:
        app_label = 'safety'

    @classmethod
    def set_django_objects(cls):
        content_type = ContentType.objects.get_or_create(app_label="safety", model="fakepost")[0]
        Permission.objects.get_or_create(codename="view_fakepost", content_type=content_type, name="Can view fake post")
        Permission.objects.get_or_create(codename="change_fakepost", content_type=content_type,
                                         name="Can change fake post")
        Permission.objects.get_or_create(codename="add_fakepost", content_type=content_type, name="Can add fake post")
        Permission.objects.get_or_create(codename="delete_fakepost", content_type=content_type,
                                         name="Can delete fake post")


@FakePost.fake_me
class TestObjectPermission(TransactionTestCase):
    def setUp(self):
        self.users = []
        self.users += [get_user_model().objects.create_user(username="TestUser", password="TestPassword")]
        self.users += [get_user_model().objects.create_user(username="TestUser2", password="TestPassword")]
        self.groups = []
        self.groups += [Group.objects.create(name="TestGroup")]
        self.groups += [Group.objects.create(name="TestGroup2")]
        self.posts = []
        self.posts += [FakePost.objects.create(title="TestPost", content="TestContent")]
        self.posts += [FakePost.objects.create(title="TestPost2", content="TestContent2")]

        FakePost.set_django_objects()

        self.fake_post_ct = ContentType.objects.get_for_model(FakePost)

    def test_set_global_permission(self):
        set_perm(self.users[0], "view_fakepost", content_type=self.fake_post_ct)

        self.assertTrue(has_perm([self.users[0]], "view_fakepost", content_type=self.fake_post_ct))

    def test_set_permission(self):
        success = set_perm(self.users[0], "view_fakepost", self.posts[0])
        self.assertTrue(success)

    def test_set_group_permission(self):
        success = set_perm(self.groups[0], "view_fakepost", self.posts[0])
        self.assertTrue(success)

    def test_has_global_permission(self):
        set_perm(self.users[0], "view_fakepost",
                 content_type=self.fake_post_ct)

        self.assertTrue(has_perm([self.users[0]], "view_fakepost", content_type=self.fake_post_ct))

    def test_has_permission(self):
        set_perm(self.users[0], "view_fakepost", self.posts[0])

        self.assertTrue(has_perm([self.users[0]], "view_fakepost", self.posts[0]))

    def test_has_permission_with_multiple_users(self):
        set_perm(self.users[0], "view_fakepost", self.posts[0])
        set_perm(self.users[1], "view_fakepost", self.posts[0])

        self.assertTrue(has_perm([self.users[0], self.users[1]], "view_fakepost", self.posts[0]))

    def test_lift_global_permission(self):
        set_perm(self.users[0], "view_fakepost", content_type=self.fake_post_ct)

        lift_perm(self.users[0], "view_fakepost", content_type=self.fake_post_ct)

        self.assertFalse(has_perm([self.users[0]], "view_fakepost"))

    def test_lift_permission(self):
        set_perm(self.users[0], "view_fakepost", self.posts[0])

        lift_perm(self.users[0], "view_fakepost", self.posts[0])

        self.assertFalse(has_perm([self.users[0]], "view_fakepost", self.posts[0]))


@FakePost.fake_me
class TestPermissionGroup(TransactionTestCase):
    def setUp(self):
        self.users = []
        self.users.append(get_user_model().objects.create_user(username="TestUser", password="TestPassword"))
        self.users.append(get_user_model().objects.create_user(username="TestUser2", password="TestPassword"))

        self.posts = []
        self.posts.append(FakePost.objects.create(title="TestPost", content="TestContent"))
        self.posts.append(FakePost.objects.create(title="TestPost2", content="TestContent2"))

        FakePost.set_django_objects()

        self.fake_post_ct = ContentType.objects.get_for_model(FakePost)

    def test_create_permission_group(self):
        create_object_group("editors",
                            ["view_fakepost", "change_fakepost", "delete_fakepost"],
                            self.posts[0])
        add_user_to_object_group(self.users[0], "editors", self.posts[0])

        self.assertTrue(has_perm([self.users[0]], "change_fakepost", self.posts[0]))

    def test_delete_permission_group(self):
        create_object_group("editors",
                            ["view_fakepost", "change_fakepost", "delete_fakepost"],
                            self.posts[0])

        delete_object_group("editors", self.posts[0])

        self.assertFalse(has_perm([self.users[0]], "change_fakepost", self.posts[0]))

    def test_remove_user_from_permission_group(self):
        create_object_group("editors",
                            ["view_fakepost", "change_fakepost", "delete_fakepost"],
                            self.posts[0])

        add_user_to_object_group(self.users[0], "editors", self.posts[0])

        remove_user_from_object_group(self.users[0], "editors", self.posts[0])

        self.assertFalse(has_perm([self.users[0]], "change_fakepost", self.posts[0]))

    def test_remove_permission_on_delete(self):
        create_object_group("editors",
                            ["view_fakepost", "change_fakepost", "delete_fakepost"],
                            self.posts[0])

        add_user_to_object_group(self.users[0], "editors", self.posts[0])

        post = self.posts[0]

        self.posts[0].delete()

        self.assertFalse(has_perm([self.users[0]], "change_fakepost", post))



from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission, Group
from django.contrib.contenttypes.models import ContentType

from safety.models import ObjectPermission, ObjectGroup


def get_object_permission_model(obj=None):
    """
    Retrieves the object permission model. If obj is provided, the Meta class of
    that object will be checked for a custom object permission model.

    Args:
        obj: The model class or instance to check for a custom object permission model.
    Returns:
        An instance of an object permission model.
    """

    # pylint: disable-next=protected-access
    # noinspection PyProtectedMember
    if obj and hasattr(obj._meta, 'object_permission_model'):
        return obj.object_permission_model

    return ContentType.objects.get_model(settings.SAFETY_OBJECT_PERMISSION_MODEL) if hasattr(settings,
                                                                                             'SAFETY_OBJECT_PERMISSION_MODEL') \
        else ObjectPermission


def get_object_group_model(obj=None):
    """
    Retrieves the object group model. If obj is provided, the Meta class of
    that object will be checked for a custom Object Group model.

    Args:
        obj: The model class or instance to check for a custom Object Group model.
    Returns:
        An instance of an Object Group model.
    """

    # pylint: disable-next=protected-access
    # noinspection PyProtectedMember
    if obj and hasattr(obj._meta, 'object_permission_model'):
        return obj.object_permission_model

    return ContentType.objects.get_model(settings.SAFETY_PERMISSION_GROUP_MODEL) if hasattr(settings,
                                                                                            'SAFETY_OBJECT_GROUP_MODEL') \
        else ObjectGroup


def has_perm(entities: list, perm: str, obj=None, content_type=None) -> bool:
    """
    Return True if the user has the specified permission. If obj is provided,
    the permission must be checked against obj. If the permission does not
    exist, return False.

    Args:
        entities: The users or groups to check the permission for.
        perm (string): The permission to check.
        obj: The object to check the permission.
        content_type (string): The content type of the permission.

    Returns:
        bool: True if the user has the specified permission, otherwise False.
    """

    all_have_perm = False

    for index, entity in enumerate(entities):
        if index == 0:
            all_have_perm = True
        if not all_have_perm or not entity.is_active:
            return False
        if entity.is_superuser:
            continue
        if not entity.is_authenticated:
            return False
        if obj is None:
            all_have_perm = entity.user_permissions.filter(codename=perm, content_type=content_type).exists()
            continue

        try:
            permission = Permission.objects.get(codename=perm)
        except Permission.DoesNotExist:
            return False

        if isinstance(entity, get_user_model()):
            all_have_perm = get_object_permission_model(obj).objects.filter(permission=permission, to_id=entity.id,
                                                                            to_ct=ContentType.objects.get_for_model(
                                                                                entity),
                                                                            object_id=obj.id).exists()
            # Check the PermissionGroup object
            if not all_have_perm:
                all_have_perm = get_object_group_model().objects.filter(target_id=obj.id,
                                                                        permissions__in=[permission],
                                                                        users__in=[entity]).exists()
        elif isinstance(entity, Group):
            all_have_perm = get_object_permission_model(obj).objects.filter(permission=permission, to_id=entity.id,
                                                                            to_ct=ContentType.objects.get_for_model(
                                                                                entity),
                                                                            object_id=obj.id).exists()

    return all_have_perm


def has_gross_perm(users: list[get_user_model()], perm: str, obj=None) -> bool:
    """
    Same as has_perm but regards groups that a user belongs to.
    Only works with users.

    Args:
        users: The users to check the permission for.
        perm (string): The permission to check.
        obj: The object to check the permission.

    Returns:
        bool: True if the user has the specified permission directly
        or through groups, otherwise False.
    """

    has_net_perm = has_perm(users, perm, obj)

    if not has_net_perm:
        return False

    for user in users:
        for group in user.groups.all():
            if has_perm(group, perm, obj):
                return True
        for group in get_object_group_model().objects.filter(users__in=[user], obj=obj):
            if has_perm(group, perm, obj):
                return True

    return False


def set_perm(entity: get_user_model() | Group, perm: str, obj: any = None, content_type: ContentType = None) -> bool:
    """
    Set a permission for a user or group.
    If the object is provided, the permission is set only on the object.
    If the object is None, the permission will be set on a model level,
    requiring the content_type argument. The content_type argument is only
    used for this purpose, it is not necessary if the object is provided
    as it will be automatically retrieved.

    Args:
        entity: The user or group to set the permission for.
        perm (string): The permission to set.
        obj: The object to set the permission on.
        content_type (ContentType): The ContentType of the object.
    """

    if obj is None:
        if content_type is None:
            raise ValueError("Content type must be provided if obj is None.")

        permission = Permission.objects.get(codename=perm, content_type=content_type)

        entity.user_permissions.add(
            permission
        )
        return True

    permission = Permission.objects.get(codename=perm, content_type=ContentType.objects.get_for_model(obj))

    if isinstance(entity, get_user_model()):
        get_object_permission_model(obj).objects.get_or_create(permission=permission, to_id=entity.id,
                                                               to_ct=ContentType.objects.get_for_model(entity),
                                                               object_id=obj.id,
                                                               object_ct=ContentType.objects.get_for_model(obj))
        return True
    elif isinstance(entity, Group):
        get_object_permission_model(obj).objects.get_or_create(permission=permission, to_id=entity.id,
                                                               to_ct=ContentType.objects.get_for_model(entity),
                                                               object_id=obj.id,
                                                               object_ct=ContentType.objects.get_for_model(obj))
        return True

    return False


def lift_perm(entity, perm: str, obj=None, content_type: ContentType = None) -> bool:
    """
    Remove the permission for a user or group.

    Args:
        entity: The user or group to remove the permission for.
        perm (string): The permission to remove.
        obj: The object to remove the permission on.
        content_type (ContentType): The ContentType of the object.
    """

    if obj is None:
        if content_type is None:
            raise ValueError("Content type must be provided if obj is None.")
        entity.user_permissions.remove(Permission.objects.get(codename=perm, content_type=content_type))
        return True

    permission = Permission.objects.get(codename=perm, content_type=ContentType.objects.get_for_model(obj))

    if type(entity) == get_user_model():
        user_obj_perm = get_object_permission_model(obj).objects.filter(permission=permission, to_id=entity.id,
                                                                        to_ct=ContentType.objects.get_for_model(
                                                                            entity),
                                                                        object_id=obj.id,
                                                                        object_ct=ContentType.objects.get_for_model(
                                                                            obj))

        if not user_obj_perm.exists():
            return False

        user_obj_perm.delete()
        return True

    group_obj_perm = get_object_permission_model(obj).objects.filter(permission=permission, to_id=entity.id,
                                                                     to_ct=ContentType.objects.get_for_model(entity),
                                                                     object_id=obj.id,
                                                                     object_ct=ContentType.objects.get_for_model(obj)
                                                                     ).delete()
    if not group_obj_perm.exists():
        return False

    group_obj_perm.delete()
    return True


def get_perm_group(name: str, obj) -> ObjectGroup:
    """
    Get a permission group.

    Args:
        name (string): The name of the group.
        obj: The object to get the group from.

    Returns:
        Group: The group object.
    """

    return get_object_group_model().objects.get(name=name, target_id=obj.id,
                                                target_ct=ContentType.objects.get_for_model(obj))


def create_object_group(name: str, permissions: list[str], obj) -> ObjectGroup:
    """
    Create a object group.

    Args:
        name (string): The name of the group.
        permissions: The permissions to add to the group.
        obj: The object to add the group to.

    Returns:
        Group: The group object.
    """

    perm_group = get_object_group_model().objects.create(name=name, target_id=obj.id,
                                                         target_ct=ContentType.objects.get_for_model(obj))

    for permission in permissions:
        perm_group.permissions.add(Permission.objects.get(codename=permission))

    return perm_group


def delete_object_group(name: str, obj) -> bool:
    """
    Remove a object group.

    Args:
        name (string): The name of the group.
        obj: The object to remove the group from.

    Returns:
        bool: True if the group was removed, otherwise False.
    """

    group = get_object_group_model().objects.filter(name=name, target_id=obj.id,
                                                    target_ct=ContentType.objects.get_for_model(obj))
    if not group.exists():
        return False

    group.delete()
    return True


def add_user_to_object_group(user: get_user_model(), name: str, obj) -> bool:
    """
    Add a user to a object group.

    Args:
        user: The user to add to the group.
        name: The name of the perm group.
        obj: The object for the perm group.

    Returns:
        bool: True if the user was added to the group, otherwise False.
    """

    get_object_group_model().objects.get(name=name, target_id=obj.id,
                                         target_ct=ContentType.objects.get_for_model(obj)).users.add(user)
    return True


def remove_user_from_object_group(user: get_user_model(), name: str, obj) -> bool:
    """
    Remove a user from a object group.

    Args:
        user: The user to remove from the group.
        name: The name of the perm group.
        obj: The object of the perm group.

    Returns:
        bool: True if the user was removed from the group, otherwise False.
    """

    user.has_perm(name, obj)

    ObjectGroup.objects.get(name=name, target_id=obj.id,
                            target_ct=ContentType.objects.get_for_model(obj)).users.remove(user)
    return True



# Generated by Django 4.2 on 2023-04-22 15:19

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('safety', '0006_alter_objectpermission_options'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='objectpermission',
            options={'verbose_name': 'User Object Permission', 'verbose_name_plural': 'User Object Permissions'},
        ),
    ]
